"""accelpy OpenAcc C++ Accelerator module"""


from accelpy.accel import AccelBase, get_compilers
from ctypes import c_int32, c_int64, c_float, c_double

t_main = """
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <stdint.h>

{varclasses}

{testcode}

{datacopies}

{kernel}

extern "C" int64_t accelpy_start() {{

    int64_t res;

    res = accelpy_kernel();

    return res;
}}

extern "C" int64_t accelpy_stop() {{

    int64_t res;

{datadeletes}

    res = 0;

    return res;
}}
"""

t_varclass = """
class {clsname} {{
public:
    {dtype} * data;

    const unsigned int size = {size};
    const unsigned int ndim = {ndim};
    const unsigned int shape[{ndim}] = {shape};
    const unsigned int stride[{ndim}] = {stride};

    {dtype} & operator() ({oparg}) {{
        return data[{offset}];
    }}

    {dtype} operator() ({oparg}) const {{
        return data[{offset}];
    }}
}};
"""

t_h2acopy = """
extern "C" int64_t {funcname}(void * data) {{
    int64_t res;

    {varname}.data = ({dtype} *) data;

    #pragma acc enter data copyin({varname})
    #pragma acc enter data copyin({varname}.data[0:{size}])

    res = 0;

    return res;
}}
"""

t_h2a_scalar = """
extern "C" int64_t {funcname}(void * data) {{
    int64_t res;

    {varname} = *({dtype} *) data;

    res = 0;

    return res;
}}
"""

t_h2amalloc = """
extern "C" int64_t {funcname}(void * data) {{
    int64_t res;


    {varname}.data = ({dtype} *) data;

    #pragma acc enter data copyin({varname})
    #pragma acc enter data create({varname}.data[0:{size}])

    res = 0;

    return res;
}}
"""

t_a2hcopy = """
extern "C" int64_t {funcname}(void * data) {{
    int64_t res;

    #pragma acc exit data copyout({varname}.data[0:{size}])

    res = 0;

    return res;
}}
"""

t_testfunc = """
extern "C" int64_t accelpy_test_run() {{
    int64_t res;
    
    #pragma acc parallel num_gangs(1) num_workers({size}) vector_length(1)
    #pragma acc loop gang worker vector
    for (int id = 0; id < {varin}.shape[0]; id++) {{
        {varout}(id) = {varin}(id);
    }}

    res = 0;

    return res;
}}
"""

t_kernel = """
extern "C" int64_t accelpy_kernel(){{

    int64_t res;

    #pragma acc parallel num_gangs(ACCELPY_TEAM_DIM0 * ACCELPY_TEAM_DIM1 * ACCELPY_TEAM_DIM2) \\
                 num_workers(ACCELPY_WORKER_DIM0 * ACCELPY_WORKER_DIM1 * ACCELPY_WORKER_DIM2) \\
               vector_length(ACCELPY_ASSIGN_DIM0 * ACCELPY_ASSIGN_DIM1 * ACCELPY_ASSIGN_DIM2)
    {{

    {order}

    }}

    res = 0;

    return res;
}}
"""

class OpenaccCppAccel(AccelBase):

    name = "openacc_cpp"
    lang = "cpp"

    # dtype: ( C type name, ctype )
    # TODO: dynamically generates on loading
    dtypemap = {
        "int32": ["int32_t", c_int32],
        "int64": ["int64_t", c_int64],
        "float32": ["float", c_float],
        "float64": ["double", c_double]
    }

    def gen_varclasses(self, inputs, outputs):

        varclasses = []
        for arg in self._testdata+inputs+outputs:

            ndim = arg["data"].ndim

            if ndim > 0:
                dtype = self.get_dtype(arg)

                oparg = ", ".join(["int dim%d"%d for d in range(ndim)])
                offset = "+".join(["stride[%d]*dim%d"%(d,d) for d in range(ndim)])

                varclasses_fmt = {
                    "clsname": "type_" + arg["curname"],
                    "size": str(arg["data"].size),
                    "ndim": str(ndim),
                    "shape": "{%s}" % self.get_shapestr(arg),
                    "stride": "{%s}" % self.get_stridestr(arg),
                    "dtype": dtype,
                    "offset":offset,
                    "oparg":oparg
                }

                varclasses.append(t_varclass.format(**varclasses_fmt))

        return "\n\n".join(varclasses)

    def gen_datacopies(self, inputs, outputs):

        out = []

        for input in inputs:
            ndim = input["data"].ndim
            dtype = self.get_dtype(input)
            funcname = self.getname_h2acopy(input)

            if ndim == 0:
                out.append("%s %s;" % (dtype, input["curname"]))

                out.append(t_h2a_scalar.format(funcname=funcname,
                        varname=input["curname"], dtype=dtype))

            else:
                out.append("type_{name} {name} = type_{name}();".format(
                        name=input["curname"]))
                out.append(t_h2acopy.format(funcname=funcname,
                        size=str(input["data"].size),
                        varname=input["curname"], dtype=dtype))

        for output in outputs:
            ndim = output["data"].ndim
            dtype = self.get_dtype(output)
            funcname = self.getname_h2amalloc(output)

            if ndim == 0:
                out.append("%s %s;" % (dtype, output["curname"]))
                out.append(t_h2a_scalar.format(funcname=funcname,
                        varname=output["curname"], dtype=dtype))

            else:
                out.append("type_{name} {name} = type_{name}();".format(
                        name=output["curname"]))
                out.append(t_h2amalloc.format(funcname=funcname, size=str(output["data"].size),
                        varname=output["curname"], dtype=dtype))

        for output in outputs:
            ndim = output["data"].ndim
            funcname = self.getname_a2hcopy(output)

            if ndim == 0:
                out.append(t_a2h.format(funcname=funcname,
                        varname=output["curname"]))

            else:
                out.append(t_a2hcopy.format(funcname=funcname, size=str(output["data"].size),
                        varname=output["curname"]))

        return "\n".join(out)

    def gen_testcode(self):

        out = []

        input = self._testdata[0]
        output = self._testdata[1]

        dtype_in = self.get_dtype(input)
        dtype_out = self.get_dtype(input)
        funcname_in = "accelpy_test_h2acopy"
        funcname_out = "accelpy_test_h2amalloc"
        funcname_a2h = "accelpy_test_a2hcopy"

        out.append("type_{name} {name} = type_{name}();".format(
                    name=input["curname"]))

        out.append("type_{name} {name} = type_{name}();".format(
                    name=output["curname"]))

        out.append(t_h2acopy.format(funcname=funcname_in, size=str(input["data"].size),
                    varname=input["curname"], dtype=dtype_in))
        out.append(t_h2amalloc.format(funcname=funcname_out, size=str(output["data"].size),
                    varname=output["curname"], dtype=dtype_out))
        out.append(t_a2hcopy.format(funcname=funcname_a2h, size=str(output["data"].size),
                    varname=output["curname"]))

        out.append(t_testfunc.format(varin=input["curname"],
                    varout=output["curname"], size=str(input["data"].size)))

        return "\n".join(out)

    def gen_kernel(self):

        order =  self._order.get_section(self.name)

        return t_kernel.format(order="\n".join(order.body))

    def gen_code(self, compiler, inputs, outputs, triple, run_id, device, channel):

        macros = {
            "ACCELPY_OPENACC_RUNID": str(run_id),
            "ACCELPY_OPENACC_NGANGS": str(triple[0][0]*triple[0][1]*triple[0][2]),
            "ACCELPY_OPENACC_NWORKERS": str(triple[1][0]*triple[1][1]*triple[1][2]),
            "ACCELPY_OPENACC_LENVECTOR": str(triple[2][0]*triple[2][1]*triple[2][2]),
        }

        datadeletes = []

        for data in inputs:
            if data["data"].ndim > 0:
                datadeletes.append("#pragma acc exit data delete(%s.data)" % data["curname"])
                datadeletes.append("#pragma acc exit data delete(%s)" % data["curname"])

        main_fmt = {
            "varclasses":self.gen_varclasses(inputs, outputs),
            "testcode":self.gen_testcode(),
            "datacopies":self.gen_datacopies(inputs, outputs),
            "kernel":self.gen_kernel(),
            "datadeletes": "\n".join(datadeletes)
        }

        code = t_main.format(**main_fmt)

        return code, macros

    def getname_h2amalloc(self, arg):
        return "accelpy_h2amalloc_%s" % arg["curname"]

    def getname_h2acopy(self, arg):
        return "accelpy_h2acopy_%s" % arg["curname"]

    def getname_a2hcopy(self, arg):
        return "accelpy_a2hcopy_%s" % arg["curname"]

OpenaccCppAccel.avails[OpenaccCppAccel.name] = OpenaccCppAccel
